import asyncio
import logging
from unittest.mock import Mock
import pytest

from aiosubpub import Channel

_LOGGER = logging.getLogger(__name__)


@pytest.fixture
def dummy_message():
    return "a dummy message"


class MockedCallback:
    def __init__(self):
        self.mock = Mock()

    def __call__(self, *args, **kwargs):
        self.mock(*args, **kwargs)


@pytest.fixture
def mock_callback():
    return MockedCallback()


@pytest.mark.asyncio
async def test_pubsub_callback_success(dummy_message, mock_callback: MockedCallback):
    dummy_channel = Channel("dummy channel")

    subscription = dummy_channel.subscribe(mock_callback)

    dummy_channel.publish(dummy_message)
    await asyncio.sleep(1)
    mock_callback.mock.assert_called_once_with(dummy_message)
    subscription.cancel()


@pytest.mark.asyncio
async def test_channel_list(mock_callback: MockedCallback):
    """Assert the subscription list is empty when starting a channel"""
    dummy_channel = Channel("dummy channel")
    assert len(dummy_channel.subscriptions) == 0


@pytest.mark.asyncio
async def test_channel_subscriptions(mock_callback):
    dummy_channel = Channel("dummy channel")

    sub1 = dummy_channel.subscribe(mock_callback)
    sub2 = dummy_channel.subscribe(mock_callback)

    assert len(dummy_channel.subscriptions) == 2

    sub1.cancel()
    sub2.cancel()


@pytest.mark.asyncio
async def test_channel_unsubscribe_no_watch(mock_callback):
    """Test subscription after unsubscribe.

    This test cancels the subscription before the watching starts.
    """
    dummy_channel = Channel("dummy channel")
    subscription_task = dummy_channel.subscribe(mock_callback)
    assert len(dummy_channel.subscriptions) == 1

    subscription_task.cancel()
    await asyncio.sleep(0.1)
    assert len(dummy_channel.subscriptions) == 0


@pytest.mark.asyncio
async def test_channel_unsubscribe_watch(mock_callback):
    """Test subscription after unsubscribe."""

    dummy_channel = Channel("dummy channel")
    subscription_task = dummy_channel.subscribe(mock_callback)
    assert len(dummy_channel.subscriptions) == 1

    await asyncio.sleep(0.1)
    subscription_task.cancel()
    await asyncio.sleep(0.1)
    assert len(dummy_channel.subscriptions) == 0


@pytest.mark.asyncio
async def test_custom_subscription(dummy_message):
    """Test a custom subscription"""

    channel = Channel("dummy channel")

    subscription = channel.get_subscription()

    async def _custom_subscriber():
        with subscription as sub:
            result = await sub.get()
            return result

    channel.publish(dummy_message)

    result = await _custom_subscriber()
    assert result == dummy_message



"""Async pubsub implementation."""

import asyncio
import logging
import random
from asyncio import CancelledError

__version__ = "1.0.3"

LOGGER = logging.getLogger(__name__)


class Channel:
    """A channel to which you can subscribe."""

    def __init__(self, name="a channel"):
        self.subscriptions = set()
        self.name = name

    def publish(self, message):
        """Publish a message on this channel."""
        for queue in self.subscriptions:
            queue.put_nowait(message)

    def get_subscription(self):
        """Return a subscription object.

        Used internally but also useful to create custom subscriber methods.
        """
        subscription = Subscription(self)
        self.subscriptions.add(subscription.queue)
        return subscription

    def subscribe(self, callback) -> "Subscription":
        """Subscribe to this channel"""
        _loop = asyncio.get_running_loop()
        subscription = self.get_subscription()
        subscription.task = _loop.create_task(self._subscribe(subscription, callback))
        return subscription

    async def _subscribe(self, subscription: "Subscription", callback):
        try:
            with subscription as queue:
                while True:
                    msg = await queue.get()
                    callback(msg)
                    queue.task_done()
        except CancelledError:
            LOGGER.debug("Shutting down subscriber")


class Subscription:
    """Subscription class.
    Used to subscribe to a Channel."""

    def __init__(self, hub):
        self.hub = hub
        self.queue = asyncio.Queue()
        self.task = None

    def cancel(self):
        """Cancel the subscription. The same as unsubscribe"""
        self.unsubscribe()

    def unsubscribe(self):
        """Unsubscribe the subscription."""
        self.task.cancel()
        self._remove_subscription()

    def _remove_subscription(self):
        if self.queue in self.hub.subscriptions:
            self.hub.subscriptions.remove(self.queue)

    def __enter__(self):
        self.hub.subscriptions.add(self.queue)
        LOGGER.debug(
            "Subscription: %s, total subscriptions %i",
            self.hub.name,
            len(self.hub.subscriptions),
        )
        return self.queue

    def __exit__(self, _type, value, traceback):
        LOGGER.debug("Un-subscribing from channel: %s", self.hub.name)
        self._remove_subscription()


async def reader(name, hub):
    """An example reader"""
    msg = ""
    with Subscription(hub) as queue:
        while msg != "SHUTDOWN":
            msg = await queue.get()
            print(f"Reader {name} got message: {msg}")

            if random.random() < 0.1:
                print(f"Reader {name} has read enough")
                break

    print(f"Reader {name} is shutting down")



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

extras_require = \
{'dev': ['pytest-cov', 'coverage[toml]>=5.0.3', 'pre-commit'],
 'test': ['pytest==5.3.5', 'pytest-asyncio==0.10.0']}

setup(name='aiosubpub',
      version='1.0.3',
      description='Async pubsub implementation.',
      author='Sander Teunissen',
      author_email='no_mail@nomail.com',
      url='https://github.com/sander76/pubsub',
      py_modules=['aiosubpub'],
      extras_require=extras_require,
      python_requires='>=3.7',
     )



import nox


def install_flit_dev_deps(session):
    session.install("flit")
    session.run("flit", "install", "--deps", "develop")


@nox.session(python=["3.7", "3.8"])
def tests(session):
    install_flit_dev_deps(session)
    session.run("pytest", "--cov=aiosubpub", "--cov-report=xml:cov.xml", "tests")
